'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) { return typeof obj; } : function (obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; };

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }(); /**
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      * @external {Parameter} https://github.com/Reading-eScience-Centre/coverage-jsapi/blob/master/Parameter.md
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      * @external {Domain} https://github.com/Reading-eScience-Centre/coverage-jsapi/blob/master/Domain.md
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      * @external {Range} https://github.com/Reading-eScience-Centre/coverage-jsapi/blob/master/Range.md
                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                      */

exports.transformParameter = transformParameter;
exports.transformDomain = transformDomain;

var _ndarray = require('ndarray');

var _ndarray2 = _interopRequireDefault(_ndarray);

var _urlTemplate = require('url-template');

var _urlTemplate2 = _interopRequireDefault(_urlTemplate);

var _covutils = require('covutils');

var _constants = require('./constants.js');

var _util = require('./util.js');

var _http = require('./http.js');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

/** 
 * Wraps a CoverageJSON Coverage object as a Coverage API object.
 * 
 * @see https://github.com/Reading-eScience-Centre/coverage-jsapi
 */
var Coverage = function () {

  /**
   * Create a Coverage instance.
   * 
   * @param {Object} covjson A CoverageJSON Coverage object.
   * @param {Object} [options] 
   * @param {boolean} [options.cacheRanges]
   *   If true, then any range that was loaded remotely is cached.
   *   (The domain is always cached.)
   * @param {Array} [options.referencing]
   *   Referencing info to use (e.g. from containing collection).                        
   */
  function Coverage(covjson, options) {
    _classCallCheck(this, Coverage);

    this._covjson = covjson;

    /**
     * The constant "Coverage".
     * 
     * @type {string}
     */
    this.type = _constants.COVERAGE;

    /**
     * JSON-LD document
     * 
     * @type {Object}
     */
    this.ld = {};

    this._exposeLd(covjson);

    this.prefixes = (0, _util.getNamespacePrefixes)(this.ld);

    /**
     * The options object that was passed in to the constructor. 
     * 
     * @type {Object} 
     */
    this.options = options ? (0, _util.shallowcopy)(options) : {};

    /** 
     * ID of the coverage.
     * 
     * @type {string|undefined} 
     */
    this.id = covjson.id;

    /**
     * A Map from key to {@link Parameter} object.
     * The key is a short alias of a {@link Parameter}, typically what is called a "variable name" or similar.
     * 
     * @type {Map<string,Parameter>}
     */
    this.parameters = new Map();
    var _iteratorNormalCompletion = true;
    var _didIteratorError = false;
    var _iteratorError = undefined;

    try {
      for (var _iterator = Object.keys(covjson.parameters)[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
        var key = _step.value;

        transformParameter(covjson.parameters, key);
        this.parameters.set(key, covjson.parameters[key]);
      }
    } catch (err) {
      _didIteratorError = true;
      _iteratorError = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion && _iterator.return) {
          _iterator.return();
        }
      } finally {
        if (_didIteratorError) {
          throw _iteratorError;
        }
      }
    }

    var domainType = void 0;
    if (typeof this._covjson.domain === 'string') {
      domainType = this._covjson.domainType;
    } else {
      domainType = this._covjson.domain.domainType || this._covjson.domainType;
    }
    if (domainType && domainType.indexOf(':') === -1) {
      domainType = _util.DOMAINTYPES_PREFIX + domainType;
    }

    /**
     * If defined, then the coverage has a domain that follows the given domain type,
     * either a full URI or a namespace-prefixed term. (See .prefixes)
     *  
     * @type {string|undefined} 
     */
    this.domainType = domainType;

    this._updateLoadStatus();
  }

  _createClass(Coverage, [{
    key: '_updateLoadStatus',
    value: function _updateLoadStatus() {
      var _this = this;

      var isLoaded = function isLoaded(prop) {
        return (typeof prop === 'undefined' ? 'undefined' : _typeof(prop)) === 'object';
      };
      var domainLoaded = isLoaded(this._covjson.domain);
      var rangesLoaded = Object.keys(this._covjson.ranges).every(function (key) {
        return isLoaded(_this._covjson.ranges[key]);
      });

      /**
       * A boolean which indicates whether all coverage data is already loaded in memory.
       * If true then this typically means that calls to .loadDomain(), .loadRange(),
       * .loadRanges(), .subsetByIndex(), and .subsetByValue() will not invoke a network request.
       * 
       * @type {boolean}
       */
      this.loaded = domainLoaded && rangesLoaded;
    }
  }, {
    key: '_exposeLd',
    value: function _exposeLd(covjson) {
      if (!covjson['@context']) {
        // no LD love here...
        return;
      }
      // make a deep copy since the object gets modified in-place later
      // but first, remove domain and range which may be embedded
      var copy = (0, _util.shallowcopy)(covjson);
      delete copy.domain;
      delete copy.ranges;
      this.ld = JSON.parse(JSON.stringify(copy));
    }

    /**
     * Returns a Promise succeeding with a {@link Domain} object.
     * 
     * @return {Promise<Domain>}
     */

  }, {
    key: 'loadDomain',
    value: function loadDomain() {
      var _this2 = this;

      var domainOrUrl = this._covjson.domain;
      if (this._domainPromise) return this._domainPromise;
      var promise = void 0;
      if ((typeof domainOrUrl === 'undefined' ? 'undefined' : _typeof(domainOrUrl)) === 'object') {
        var domain = domainOrUrl;
        transformDomain(domain, this.options.referencing, this.domainType);
        promise = Promise.resolve(domain);
      } else {
        var url = domainOrUrl;
        promise = (0, _http.load)(url).then(function (result) {
          var domain = result.data;
          transformDomain(domain, _this2.options.referencing, _this2.domainType);
          _this2._covjson.domain = domain;
          _this2._updateLoadStatus();
          return domain;
        });
      }
      /* The promise gets cached so that the domain is not loaded twice remotely.
       * This might otherwise happen when loadDomain and loadRange is used
       * with Promise.all(). Remember that loadRange also invokes loadDomain.
       */
      this._domainPromise = promise;
      return promise;
    }

    /**
     * Returns a Promise succeeding with a {@link Range} object.
     * 
     * Note that this method implicitly loads the domain as well. 
     * 
     * @example
     * cov.loadRange('salinity').then(function (sal) {
     *   // work with Range object
     * }).catch(function (e) {
     *   // there was an error when loading the range
     *   console.log(e.message)
     * }) 
     * @param {string} paramKey The key of the Parameter for which to load the range.
     * @return {Promise<Range>} A Promise object which loads the requested range data and succeeds with a Range object.
     */

  }, {
    key: 'loadRange',
    value: function loadRange(paramKey) {
      return loadRangeFn(this)(paramKey);
    }

    /**
     * Returns the requested range data as a Promise.
     * 
     * Note that this method implicitly loads the domain as well. 
     * 
     * @example
     * cov.loadRanges(['salinity','temp']).then(function (ranges) {
     *   // work with Map object
     *   console.log(ranges.get('salinity').values)
     * }).catch(function (e) {
     *   // there was an error when loading the range data
     *   console.log(e)
     * }) 
     * @param {iterable<string>} [paramKeys] An iterable of parameter keys for which to load the range data. If not given, loads all range data.
     * @return {Promise<Map<string,Range>>} A Promise object which loads the requested range data and succeeds with a Map object.
     */

  }, {
    key: 'loadRanges',
    value: function loadRanges(paramKeys) {
      return loadRangesFn(this)(paramKeys);
    }

    /**
     * Returns a Promise object which provides a copy of this Coverage object
     * with the domain subsetted by the given indices specification.
     * 
     * Note that the coverage type and/or domain type of the resulting coverage
     * may be different than in the original coverage.
     * 
     * Note that the subsetted ranges are a view over the original ranges, meaning
     * that no copying is done but also no memory is released if the original
     * coverage is garbage collected.
     * 
     * @example
     * cov.subsetByIndex({t: 4, z: {start: 10, stop: 20} }).then(function(subsetCov) {
     *   // work with subsetted coverage
     * })
     * @param {Object} constraints An object which describes the subsetting constraints.
     *   Every property of it refers to an axis name as defined in Domain.names,
     *   and its value must either be an integer
     *   or an object with start, stop, and optionally step (defaults to 1) properties
     *   whose values are integers.
     *   Properties that have the values undefined or null are ignored. 
     *   All integers must be non-negative, step must not be zero.
     *   An integer constrains the axis to the given index,
     *   a start/stop/step object to a range of indices:
     *   If step=1, this includes all indices starting at start and ending at stop (exclusive);
     *   if step>1, all indices start, start + step, ..., start + (q + r - 1) step where 
     *   q and r are the quotient and remainder obtained by dividing stop - start by step.
     * @returns {Promise<Coverage>} A Promise object with the subsetted coverage object as result.
     */

  }, {
    key: 'subsetByIndex',
    value: function subsetByIndex(constraints) {
      return subsetByIndexFn(this)(constraints);
    }

    /**
     * Returns a Promise object which provides a copy of this Coverage object
     * with the domain subsetted by the given value specification.
     * 
     * Note that the coverage type and/or domain type of the resulting coverage
     * may be different than in the original coverage.
     * 
     * Note that the subsetted ranges are a view over the original ranges, meaning
     * that no copying is done but also no memory is released if the original
     * coverage is garbage collected.
     * 
     * @example
     * cov.subsetByValue({
     *   t: '2015-01-01T01:00:00',
     *   z: {start: -10, stop: -5} 
     * }).then(function(subsetCov) {
     *   // work with subsetted coverage
     * })
     * @example
     * cov.subsetByValue({z: {target: -10} }).then(function(subsetCov) {
     *   // work with subsetted coverage
     * }
     * @param {Object} constraints An object which describes the subsetting constraints.
     *  Every property of it refers to an axis name as defined in Domain.names,
     *  and its value must either be a number or string, or,
     *  if the axis has an ordering relation, an object with start and stop properties
     *  whose values are numbers or strings, or an object with a target property
     *  whose value is a number or string.
     *  Properties that have the values undefined or null are ignored.
     *  A number or string constrains the axis to exactly the given value,
     *  a start/stop object to the values intersecting the extent,
     *  and a target object to the value closest to the given value.
     * @returns {Promise<Coverage>} A Promise object with the subsetted coverage object as result.
     */

  }, {
    key: 'subsetByValue',
    value: function subsetByValue(constraints) {
      return (0, _covutils.subsetByValue)(this, constraints);
    }
  }]);

  return Coverage;
}();

exports.default = Coverage;


function getRangeAxisOrder(domain, range) {
  if (!domain) {
    return range._axisNames;
  }
  // backwards-compatibility, in the future the range always has an explicit axis ordering
  var needsRangeAxisOrder = [].concat(_toConsumableArray(domain.axes.values())).filter(function (axis) {
    return axis.values.length > 1;
  }).length > 1;

  // domain is checked for backwards-compatibility
  var axisOrder = domain._rangeAxisOrder || range._axisNames;
  if (needsRangeAxisOrder && !axisOrder) {
    throw new Error('Range axis order missing');
  }
  axisOrder = axisOrder || [].concat(_toConsumableArray(domain.axes.keys()));
  return axisOrder;
}

function getRangeShapeArray(domain, range) {
  if (!domain) {
    return range._shape;
  }
  // mostly backwards-compatibility, in the future this just returns range._shape
  var axisOrder = getRangeAxisOrder(domain, range);
  var shape = axisOrder.map(function (k) {
    return domain.axes.get(k).values.length;
  });
  if (range._shape) {
    var matchesDomain = range._shape.length === shape.length && range._shape.every(function (v, i) {
      return v === shape[i];
    });
    if (!matchesDomain) {
      throw new Error('range.shape must match domain axis sizes');
    }
  }
  return shape;
}

function loadRangesFn(cov) {
  return function (paramKeys) {
    if (paramKeys === undefined) paramKeys = cov.parameters.keys();
    paramKeys = Array.from(paramKeys);
    return Promise.all(paramKeys.map(function (k) {
      return cov.loadRange(k);
    })).then(function (ranges) {
      var map = new Map();
      for (var i = 0; i < paramKeys.length; i++) {
        map.set(paramKeys[i], ranges[i]);
      }
      return map;
    });
  };
}

function loadRangeFn(cov, globalConstraints) {
  return function (paramKey) {
    // Since the shape of the range array is derived from the domain, it has to be loaded as well.
    return cov.loadDomain().then(function () {
      var rangeOrUrl = cov._covjson.ranges[paramKey];
      if ((typeof rangeOrUrl === 'undefined' ? 'undefined' : _typeof(rangeOrUrl)) === 'object') {
        var rawRange = rangeOrUrl;
        // we need the original domain here, not a potentially subsetted one,
        // therefore we access cov._covjson directly
        // this legacy code will disappear once the old range format is not supported anymore
        return doLoadRange(cov, paramKey, rawRange, cov._covjson.domain, globalConstraints);
      } else {
        var url = rangeOrUrl;
        return (0, _http.load)(url).then(function (result) {
          var rawRange = result.data;
          return doLoadRange(cov, paramKey, rawRange, cov._covjson.domain, globalConstraints);
        });
      }
    });
  };
}

function doLoadRange(cov, paramKey, range, domain) {
  var globalConstraints = arguments.length > 4 && arguments[4] !== undefined ? arguments[4] : {};

  globalConstraints = (0, _covutils.normalizeIndexSubsetConstraints)(domain, globalConstraints);

  if (range.type === 'NdArray' || range.type === 'Range') {
    // if an NdArray, then we modify it in-place (only done the first time)
    transformNdArrayRange(range, domain);
    if (cov.options.cacheRanges) {
      cov._covjson.ranges[paramKey] = range;
      cov._updateLoadStatus();
    }

    var newrange = subsetNdArrayRangeByIndex(range, domain, globalConstraints);
    return Promise.resolve(newrange);
  } else if (range.type === 'TiledNdArray') {
    return loadTiledNdArraySubset(range, globalConstraints);
  } else {
    throw new Error('Unsupported: ' + range.type);
  }
}

/**
 * 
 * @param {object} range TiledNdArray range object
 * @param {object} constraints subsetting constraints
 * @returns {Promise<Range>}
 */
function loadTiledNdArraySubset(range, constraints) {
  var constraintsArr = range.axisNames.map(function (name) {
    return constraints[name];
  });

  // step 1: select tileset with least network effort
  var fillNulls = function fillNulls(tileShape) {
    return tileShape.map(function (v, i) {
      return v === null ? range.shape[i] : v;
    });
  };
  var tilesetsStats = range.tileSets.map(function (ts) {
    return getTilesetStats(fillNulls(ts.tileShape), constraintsArr);
  });
  var idxBestTileset = indexOfBestTileset(tilesetsStats);
  var tileset = range.tileSets[idxBestTileset];
  var urlTemplate = _urlTemplate2.default.parse(tileset.urlTemplate);
  var tileShape = fillNulls(tileset.tileShape);

  // step 2: determine the tiles to load
  var subsetTilesetAxes = [];
  for (var ax = 0; ax < tileShape.length; ax++) {
    var _constraintsArr$ax = constraintsArr[ax];
    var start = _constraintsArr$ax.start;
    var stop = _constraintsArr$ax.stop;
    var step = _constraintsArr$ax.step;

    var tileSize = tileShape[ax];

    // the indices of the first and last tile containing the subsetting constraints
    var tileStart = Math.floor(start / tileSize); // inclusive
    var tileStop = Math.ceil(stop / tileSize); // exclusive

    var tilesetAxis = [];
    for (var t = tileStart; t < tileStop; t++) {
      var mid = (t + 0.5) * tileSize;
      // regard the subset constraint as a list of [x,y) half-closed intervals and find out where 'mid' falls into
      var iv = Math.floor((mid - start) / step);

      // start and end point of the interval in range index space
      var ivStart = start + iv * step;
      var ivStop = start + (iv + 1) * step;

      // tile start and end in range index space
      var tileStartR = t * tileSize;
      var tileStopR = (t + 1) * tileSize;

      // check if the start or end point of the interval lies within the tile
      if (ivStart >= tileStartR || tileStopR <= ivStop) {
        tilesetAxis.push(t);
      }
    }
    subsetTilesetAxes.push(tilesetAxis);
  }

  var tiles = cartesianProduct(subsetTilesetAxes);
  var subsetShape = constraintsArr.map(function (_ref) {
    var start = _ref.start;
    var stop = _ref.stop;
    var step = _ref.step;
    return Math.floor((stop - start) / step) + (stop - start) % step;
  });

  function getTileUrl(tile) {
    var tileUrlVars = {};
    tile.forEach(function (v, i) {
      return tileUrlVars[range.axisNames[i]] = v;
    });
    return urlTemplate.expand(tileUrlVars);
  }

  // step 3a: check if only a single tile will be loaded and avoid copying data around in that case
  if (tiles.length === 1 && subsetShape.every(function (v, i) {
    return v === tileShape[i];
  })) {
    var url = getTileUrl(tiles[0]);
    return (0, _http.load)(url).then(function (result) {
      var tileRange = result.data;
      transformNdArrayRange(tileRange);
      return tileRange;
    });
  } else {
    var _ret = function () {
      // step 3b: create an empty ndarray of the subset shape that will be filled with tile data
      var subsetSize = subsetShape.reduce(function (l, r) {
        return l * r;
      });
      var subsetNdArr = (0, _ndarray2.default)(new Array(subsetSize), subsetShape);

      // step 4: load tiles and fill subset ndarray
      var promises = tiles.map(function (tile) {
        var url = getTileUrl(tile);
        return (0, _http.load)(url).then(function (result) {
          var tileRange = result.data;
          transformNdArrayRange(tileRange);

          // figure out which parts of the tile to copy into which part of the final ndarray
          var tileOffsets = tile.map(function (v, i) {
            return v * tileShape[i];
          });

          // iterate all tile values and for each check if they are part of the subset
          // TODO this code is probably quite slow, consider pre-compiling etc
          //      -> use ndarray-ops#assign for fast copies
          var tileAxesSubsetIndices = [];
          for (var _ax = 0; _ax < tileShape.length; _ax++) {
            var _constraintsArr$_ax = constraintsArr[_ax];
            var start = _constraintsArr$_ax.start;
            var stop = _constraintsArr$_ax.stop;
            var step = _constraintsArr$_ax.step;

            var tileAxisSize = tileShape[_ax];
            var tileAxisOffset = tileOffsets[_ax];
            var tileAxisSubsetIndices = [];
            var startIdx = 0;
            if (tileAxisOffset < start) {
              startIdx = start - tileAxisOffset;
            }
            var stopIdx = tileAxisSize;
            if (tileAxisOffset + stopIdx > stop) {
              stopIdx = stop - tileAxisOffset;
            }

            for (var i = startIdx; i < stopIdx; i++) {
              var idx = tileAxisOffset + i;
              if ((idx - start) % step === 0) {
                tileAxisSubsetIndices.push(i);
              }
            }
            tileAxesSubsetIndices.push(tileAxisSubsetIndices);
          }
          var tileSubsetIndices = cartesianProduct(tileAxesSubsetIndices);
          var _iteratorNormalCompletion2 = true;
          var _didIteratorError2 = false;
          var _iteratorError2 = undefined;

          try {
            for (var _iterator2 = tileSubsetIndices[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
              var _tileRange$_ndarr;

              var tileInd = _step2.value;

              var val = (_tileRange$_ndarr = tileRange._ndarr).get.apply(_tileRange$_ndarr, _toConsumableArray(tileInd));
              var subsetInd = tileInd.map(function (i, ax) {
                var idx = tileOffsets[ax] + i;
                return Math.floor((idx - constraintsArr[ax].start) / constraintsArr[ax].step);
              });
              subsetNdArr.set.apply(subsetNdArr, _toConsumableArray(subsetInd).concat([val]));
            }
          } catch (err) {
            _didIteratorError2 = true;
            _iteratorError2 = err;
          } finally {
            try {
              if (!_iteratorNormalCompletion2 && _iterator2.return) {
                _iterator2.return();
              }
            } finally {
              if (_didIteratorError2) {
                throw _iteratorError2;
              }
            }
          }
        });
      });

      // step 5: create and return the new range
      return {
        v: Promise.all(promises).then(function () {
          var newrange = {
            dataType: range.dataType,
            get: createRangeGetFunction(subsetNdArr, range.axisNames),
            _ndarr: subsetNdArr,
            _axisNames: range.axisNames,
            _shape: subsetShape
          };
          newrange.shape = new Map(range.axisNames.map(function (v, i) {
            return [v, subsetNdArr.shape[i]];
          }));
          return newrange;
        })
      };
    }();

    if ((typeof _ret === 'undefined' ? 'undefined' : _typeof(_ret)) === "object") return _ret.v;
  }
}

/**
 * Return the cartesian product of the given arrays.
 * 
 * @see http://stackoverflow.com/a/36234242
 */
function cartesianProduct(arr) {
  return arr.reduce(function (a, b) {
    return a.map(function (x) {
      return b.map(function (y) {
        return x.concat(y);
      });
    }).reduce(function (a, b) {
      return a.concat(b);
    }, []);
  }, [[]]);
}

/**
 * Returns the number of tiles and values that have to be loaded, given a set of subsetting constraints.
 * 
 * @param {Array<number>} tileShape
 * @param {Array<object>} constraints - start/stop/step subset constraints for each axis, stop is exclusive
 * @returns {number}
 */
function getTilesetStats(tileShape, constraints) {
  var tileCount = 1;
  for (var i = 0; i < tileShape.length; i++) {
    var _constraints$i = constraints[i];
    var start = _constraints$i.start;
    var stop = _constraints$i.stop;
    var step = _constraints$i.step;

    var tileSize = tileShape[i];

    // the indices of the first and last tile containing the subsetting constraints
    var tileStart = Math.floor(start / tileSize); // inclusive
    var tileStop = Math.ceil(stop / tileSize); // exclusive

    // total number of values within the tiles containing the subsetting constraints
    var nvalues = tileSize * (tileStop - tileStart);

    // number of tiles that intersect with the subsetting constraints
    tileCount *= Math.ceil(nvalues / Math.max(step, tileSize));
  }
  // the value count is an upper bound as it doesn't account for edge tiles that may be smaller
  var valueCount = tileCount * tileShape.reduce(function (l, r) {
    return l * r;
  });

  return { tileCount: tileCount, valueCount: valueCount };
}

/**
 * Returns the index of the tileset with minimum network effort based on the given tileset statistics.
 * Effort here means a combination of number of requested tiles and values.
 * 
 * @param {Array<object>} tilesetsStats
 * @returns {number} index of the tileset with minimum network effort
 */
function indexOfBestTileset(tilesetsStats) {
  // one tile request shall have an equal effort as receiving 1000 values
  var tileValueRatio = 1000;
  var efforts = tilesetsStats.map(function (s) {
    return s.tileCount + s.valueCount / tileValueRatio;
  });
  var minEffortIdx = efforts.reduce(function (imin, x, i, arr) {
    return x < arr[imin] ? i : imin;
  }, 0);
  return minEffortIdx;
}

function subsetNdArrayRangeByIndex(range, domain, constraints) {
  var _ndarr$hi$lo, _ndarr$hi;

  var ndarr = range._ndarr;

  // fast ndarray view
  var axisNames = getRangeAxisOrder(domain, range);
  var los = axisNames.map(function (name) {
    return constraints[name].start;
  });
  var his = axisNames.map(function (name) {
    return constraints[name].stop;
  });
  var steps = axisNames.map(function (name) {
    return constraints[name].step;
  });
  var newndarr = (_ndarr$hi$lo = (_ndarr$hi = ndarr.hi.apply(ndarr, _toConsumableArray(his))).lo.apply(_ndarr$hi, _toConsumableArray(los))).step.apply(_ndarr$hi$lo, _toConsumableArray(steps));

  var newrange = {
    dataType: range.dataType,
    get: createRangeGetFunction(newndarr, axisNames),
    _ndarr: newndarr,
    _axisNames: axisNames,
    _shape: newndarr.shape
  };
  newrange.shape = new Map(axisNames.map(function (v, i) {
    return [v, newndarr.shape[i]];
  }));
  return newrange;
}

function subsetByIndexFn(cov, globalConstraints) {
  return function (constraints) {
    return cov.loadDomain().then(function (domain) {
      constraints = (0, _covutils.normalizeIndexSubsetConstraints)(domain, constraints);
      var newdomain = (0, _covutils.subsetDomainByIndex)(domain, constraints);

      var newGlobalConstraints = toGlobalSubsetConstraints(constraints, globalConstraints);

      // backwards-compatibility
      if (domain._rangeAxisOrder) {
        newdomain._rangeAxisOrder = domain._rangeAxisOrder;
      }

      // assemble everything to a new coverage
      var newcov = {
        _covjson: cov._covjson,
        options: cov.options,
        type: _constants.COVERAGE,
        // TODO are the profiles still valid?
        domainProfiles: cov.domainProfiles,
        domainType: cov.domainType,
        parameters: cov.parameters,
        loadDomain: function loadDomain() {
          return Promise.resolve(newdomain);
        }
      };
      newcov.loadRange = loadRangeFn(newcov, newGlobalConstraints);
      newcov.loadRanges = loadRangesFn(newcov);
      newcov.subsetByIndex = subsetByIndexFn(newcov, newGlobalConstraints);
      newcov.subsetByValue = _covutils.subsetByValue.bind(null, newcov);
      return newcov;
    });
  };
}

/**
 * Currently unused, but may need in future.
 * This determines the best array type for categorical data which
 * doesn't have missing values.
 */
/*
function arrayType (validMin, validMax) {
  let type
  if (validMin !== undefined) {
    if (validMin >= 0) {
      if (validMax < Math.pow(2,8)) {
        type = Uint8Array
      } else if (validMax < Math.pow(2,16)) {
        type = Uint16Array
      } else if (validMax < Math.pow(2,32)) {
        type = Uint32Array
      } else {
        type = Array
      }
    } else {
      let max = Math.max(Math.abs(validMin), validMax)
      if (max < Math.pow(2,8)) {
        type = Int8Array
      } else if (validMax < Math.pow(2,16)) {
        type = Int16Array
      } else if (validMax < Math.pow(2,32)) {
        type = Int32Array
      } else {
        type = Array
      }
    }
  } else {
    type = Array
  }
  return type
}
*/

/**
 * Transforms a CoverageJSON parameter to the Coverage API format, that is,
 * some elements are converted from objects to Maps. Transformation is made in-place.
 * 
 * @param {Object} param The original parameter.
 * @access private
 */
function transformParameter(params, key) {
  if ('__transformDone' in params[key]) return;
  var param = params[key];
  param.key = key;
  if (param.unit) {
    if (typeof param.unit.symbol === 'string') {
      param.unit.symbol = {
        value: param.unit.symbol
        // no type member, since the scheme is unknown
      };
    }
  }
  if (param.categoryEncoding) {
    var map = new Map();
    var _iteratorNormalCompletion3 = true;
    var _didIteratorError3 = false;
    var _iteratorError3 = undefined;

    try {
      for (var _iterator3 = Object.keys(param.categoryEncoding)[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
        var category = _step3.value;

        var vals = param.categoryEncoding[category];
        if (!Array.isArray(vals)) {
          vals = [vals];
        }
        map.set(category, vals);
      }
    } catch (err) {
      _didIteratorError3 = true;
      _iteratorError3 = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion3 && _iterator3.return) {
          _iterator3.return();
        }
      } finally {
        if (_didIteratorError3) {
          throw _iteratorError3;
        }
      }
    }

    param.categoryEncoding = map;
  }
  param.__transformDone = true;
}

/**
 * Transforms a CoverageJSON NdArray range to the Coverage API format. Transformation is made in-place.
 * 
 * @param {Object} range The original NdArray range.
 * @param {Object} [domain] The CoverageJSON domain object. 
 * @return {Object} The transformed range.
 */
function transformNdArrayRange(range, domain) {
  if ('__transformDone' in range) return;

  var values = range.values;

  if (range.actualMin === undefined) {
    var _minMax = (0, _covutils.minMax)(values);

    var _minMax2 = _slicedToArray(_minMax, 2);

    var min = _minMax2[0];
    var max = _minMax2[1];

    if (min !== null) {
      range.actualMin = min;
      range.actualMax = max;
    }
  }

  // store the array as we will expose a Map on range.shape in the next step
  if (range.shape) {
    range._shape = range.shape;
  }
  if (range.axisNames) {
    // not part of public API
    range._axisNames = range.axisNames;
    delete range.axisNames;
  }

  var axisNames = getRangeAxisOrder(domain, range);
  var shapeArr = getRangeShapeArray(domain, range);

  var ndarr = (0, _ndarray2.default)(values, shapeArr);
  range._ndarr = ndarr;
  range.get = createRangeGetFunction(ndarr, axisNames);
  range.shape = new Map(axisNames.map(function (v, i) {
    return [v, shapeArr[i]];
  }));

  range.__transformDone = true;
  return range;
}

/**
 * 
 * @param axisOrder An array of axis names.
 * @returns Function
 */
function createRangeGetFunction(ndarr, axisOrder) {
  // see below for slower reference version
  var ndargs = '';
  for (var i = 0; i < axisOrder.length; i++) {
    if (ndargs) ndargs += ',';
    ndargs += '\'' + axisOrder[i] + '\' in obj ? obj[\'' + axisOrder[i] + '\'] : 0';
  }
  var fn = new Function('ndarr', 'return function ndarrget (obj) { return ndarr.get(' + ndargs + ') }')(ndarr);
  return fn;
}

/*
 * Reference version of createRangeGetFunction().
 * Around 50% slower (on Chrome 46) compared to precompiled version.
 * 
function createRangeGetFunction (ndarr, axisOrder) {
  axisOrder = axisOrder.slice() // help the JIT (possibly..)
  const axisCount = axisOrder.length
  return obj => {
    let indices = new Array(axisCount)
    for (let i=0; i < axisCount; i++) {
      indices[i] = axisOrder[i] in obj ? obj[axisOrder[i]] : 0
    }
    return ndarr.get(...indices)
  }
}
*/

/**
 * Transforms a CoverageJSON domain to the Coverage API format.
 * Transformation is made in-place.
 * 
 * @param {Object} domain The original domain object.
 * @param {Array} [referencing] Referencing info to inject.
 * @return {Object} The transformed domain object.
 * @access private
 */
function transformDomain(domain, referencing, domainType) {
  if ('__transformDone' in domain) return;

  domainType = domain.domainType || domainType;
  if (domainType && domainType.indexOf(':') === -1) {
    domainType = _util.DOMAINTYPES_PREFIX + domainType;
  }
  domain.domainType = domainType;

  var axes = new Map(); // axis name -> axis object

  var _iteratorNormalCompletion4 = true;
  var _didIteratorError4 = false;
  var _iteratorError4 = undefined;

  try {
    for (var _iterator4 = Object.keys(domain.axes)[Symbol.iterator](), _step4; !(_iteratorNormalCompletion4 = (_step4 = _iterator4.next()).done); _iteratorNormalCompletion4 = true) {
      var axisName = _step4.value;

      axes.set(axisName, domain.axes[axisName]);
    }
  } catch (err) {
    _didIteratorError4 = true;
    _iteratorError4 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion4 && _iterator4.return) {
        _iterator4.return();
      }
    } finally {
      if (_didIteratorError4) {
        throw _iteratorError4;
      }
    }
  }

  domain.axes = axes;

  // expand start/stop/num regular axes
  // replace 1D numeric axis arrays with typed arrays for efficiency
  var _iteratorNormalCompletion5 = true;
  var _didIteratorError5 = false;
  var _iteratorError5 = undefined;

  try {
    for (var _iterator5 = axes[Symbol.iterator](), _step5; !(_iteratorNormalCompletion5 = (_step5 = _iterator5.next()).done); _iteratorNormalCompletion5 = true) {
      var _step5$value = _slicedToArray(_step5.value, 2);

      var key = _step5$value[0];
      var axis = _step5$value[1];

      axis.key = key;

      if (axis.dataType && axis.dataType.indexOf(':') === -1) {
        axis.dataType = _util.CORE_PREFIX + axis.dataType;
      }

      // TODO remove this if-block later, just here for backwards-compatibility 
      if (axis.components) {
        axis.coordinates = axis.components;
      }

      if (!axis.coordinates) {
        axis.coordinates = [key];
      }

      // TODO remove this line later, just here for backwards-compatibility 
      axis.components = axis.coordinates;

      if ('start' in axis && 'stop' in axis && 'num' in axis) {
        var arr = new Float64Array(axis.num);
        var step = void 0;
        if (axis.num === 1) {
          if (axis.start !== axis.stop) {
            throw new Error('regular axis of length 1 must have equal start/stop values');
          }
          step = 0;
        } else {
          step = (axis.stop - axis.start) / (axis.num - 1);
        }
        for (var i = 0; i < axis.num; i++) {
          arr[i] = axis.start + i * step;
        }

        axis.values = arr;
        delete axis.start;
        delete axis.stop;
        delete axis.num;
      }

      if (Array.isArray(axis.values) && typeof axis.values[0] === 'number') {
        var _arr = new Float64Array(axis.values.length);
        for (var _i = 0; _i < axis.values.length; _i++) {
          _arr[_i] = axis.values[_i];
        }
        axis.values = _arr;
      }

      axis.bounds = wrapBounds(axis);
    }
  } catch (err) {
    _didIteratorError5 = true;
    _iteratorError5 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion5 && _iterator5.return) {
        _iterator5.return();
      }
    } finally {
      if (_didIteratorError5) {
        throw _iteratorError5;
      }
    }
  }

  if (referencing) {
    domain.referencing = referencing;
  }

  // TODO remove this later, just here for backwards-compatibility 
  var _iteratorNormalCompletion6 = true;
  var _didIteratorError6 = false;
  var _iteratorError6 = undefined;

  try {
    for (var _iterator6 = domain.referencing[Symbol.iterator](), _step6; !(_iteratorNormalCompletion6 = (_step6 = _iterator6.next()).done); _iteratorNormalCompletion6 = true) {
      var obj = _step6.value;

      if (obj.components) {
        obj.coordinates = obj.components;
      }
    }
  } catch (err) {
    _didIteratorError6 = true;
    _iteratorError6 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion6 && _iterator6.return) {
        _iterator6.return();
      }
    } finally {
      if (_didIteratorError6) {
        throw _iteratorError6;
      }
    }
  }

  if (domain.rangeAxisOrder) {
    domain._rangeAxisOrder = domain.rangeAxisOrder;
    delete domain.rangeAxisOrder;
  }

  domain.__transformDone = true;

  return domain;
}

/**
 * Applies the local index subset constraints to the existing global constraints.
 * Both constraint objects must be normalized, that is, must contain the same axes
 * as start/stop/step objects.
 * 
 * @example
 * var local = {x: {start: 0, stop: 50, step: 2}}
 * var global = {x: {start: 500, stop: 1000}}
 * var newGlobal = toGlobalSubsetConstraints(local, global)
 * // newGlobal == {x: {start: 500, stop: 550, step: 2}}
 * 
 * @example
 * var local = {x: {start: 5, stop: 10, step: 2}} // 5, 7, 9
 * var global = {x: {start: 500, stop: 1000, step: 10}} // 500, 510, 520,...
 * var newGlobal = toGlobalSubsetConstraints(local, global) 
 * // newGlobal == {x: {start: 550, stop: 600, step: 20}} // 550, 570, 590
 */
function toGlobalSubsetConstraints(localConstraints) {
  var globalConstraints = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

  var res = {};
  var _iteratorNormalCompletion7 = true;
  var _didIteratorError7 = false;
  var _iteratorError7 = undefined;

  try {
    for (var _iterator7 = Object.keys(localConstraints)[Symbol.iterator](), _step7; !(_iteratorNormalCompletion7 = (_step7 = _iterator7.next()).done); _iteratorNormalCompletion7 = true) {
      var axis = _step7.value;

      var local = localConstraints[axis];

      var _ref2 = globalConstraints[axis] || {};

      var _ref2$start = _ref2.start;
      var globalStart = _ref2$start === undefined ? 0 : _ref2$start;
      var _ref2$step = _ref2.step;
      var globalStep = _ref2$step === undefined ? 1 : _ref2$step;

      res[axis] = {
        start: globalStart + globalStep * local.start,
        stop: globalStart + globalStep * local.stop,
        step: globalStep * local.step
      };
    }
  } catch (err) {
    _didIteratorError7 = true;
    _iteratorError7 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion7 && _iterator7.return) {
        _iterator7.return();
      }
    } finally {
      if (_didIteratorError7) {
        throw _iteratorError7;
      }
    }
  }

  return res;
}

function wrapBounds(axis) {
  if (axis.bounds) {
    var _ret2 = function () {
      var bounds = axis.bounds;
      return {
        v: {
          get: function get(i) {
            return [bounds[2 * i], bounds[2 * i + 1]];
          }
        }
      };
    }();

    if ((typeof _ret2 === 'undefined' ? 'undefined' : _typeof(_ret2)) === "object") return _ret2.v;
  }
}