'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.COVJSON_PROFILE_STANDALONE = exports.EXT = exports.MEDIATYPE = undefined;
exports.getAcceptHeader = getAcceptHeader;
exports.matchesMediaTypes = matchesMediaTypes;

var _util = require('./util.js');

var MEDIATYPE = exports.MEDIATYPE = {
  COVJSON: 'application/prs.coverage+json',
  JSONLD: 'application/ld+json',
  JSON: 'application/json',
  TEXT: 'text/plain'
};

var EXT = exports.EXT = {
  COVJSON: '.covjson'
};

var COVJSON_PROFILE_STANDALONE = exports.COVJSON_PROFILE_STANDALONE = _util.CORE_PREFIX + 'standalone';

/**
 * Returns an Accept header value for requesting CoverageJSON documents.
 * 
 * @param {bool} standalone Whether to include the standalone profile of CoverageJSON or not.
 */
function getAcceptHeader(standalone) {
  var covjsonProfile = standalone ? '; profile="' + COVJSON_PROFILE_STANDALONE + '"' : '';
  var accept = MEDIATYPE.COVJSON + covjsonProfile + '; q=1.0, ' + MEDIATYPE.JSONLD + '; q=0.1, ' + MEDIATYPE.JSON + '; q=0.1';
  return accept;
}

/**
 * Checks if a media type matches any given media types, ignoring any parameters. 
 * 
 * @param {string} mediaType The media type.
 * @param {string|Array} matchingMediaTypes The media type(s) to match against.
 * @return {bool} True if there is a match.
 */
function matchesMediaTypes(mediaType, matchingMediaTypes) {
  if (!Array.isArray(matchingMediaTypes)) {
    matchingMediaTypes = [matchingMediaTypes];
  }
  return matchingMediaTypes.some(function (t) {
    return mediaType.indexOf(t) === 0;
  });
}